%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% create the tree structure.
% 
% input:
%   ntotal = 2^L: the dimension of the problem.
%   NODES(:,j): stored information for each node.
%       1 - level, 
%       2 - id within level, starting from 1. For leaf node, this tells
%           which x-variable to integrate.
%       3 - =1,2,3,or 4, boundary or interior node.
%           0: root node.
%           1: left boundary, 2: right boundary
%           3: left interior, 4: right interior
%       4 - parent, 
%       5~6 - children, 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

clear variables; close all; clc; format long e;

% global variables.
%   NODES: tree structure
%   COEFF: the function coefficients for each node point.
%   ABvec(1:2,ntotal): the integration bound for each variable. 
%   FourierInfo: the number of terms (FourierInfo.num)
%       and the fundamental period (FourierInfo.LPeriod, function is 
%       periodic in [-Lperiod, Lperiod].
%   XLoc: The evenly distributed sample points in [-Lperiod, Lperiod].
%   FilterFun(1:FourierInfo.num*2, 1:FourierInfo.num*2): the filter
%       function values, so the function is well-approximated in 
%       [-LPeriod/2,LPeriod/2], and decays to 0 smoothly outside.
global NODES COEFF ABvec FourierInfo XLoc FilterFun

% test 2^ntotal_test cases.
ntotal_test(1) = 11;
ntotal = 2^ntotal_test(1)
FourierInfo.num=256;FourierInfo.LPeriod=14;
NODES = zeros(6,2*ntotal-1); % the tree structure.
COEFF = zeros(2*FourierInfo.num,2*FourierInfo.num,2*ntotal-1); % Fourier coefficients for each node.
ABvec=zeros(2,ntotal); 
ABvec(1,:)=-ones(1,ntotal); ABvec(2,:)=ones(1,ntotal);
ABvec(2,1)=0.5;ABvec(2,2)=2; % no symmetric example, two points, compare with Mathematica to show 16 digits.

% construct the filter function using error function.
h=2/FourierInfo.num; eps=1/14;
x=-2:h:2-h/2; XLoc=x/2*FourierInfo.LPeriod;
FilterFun=(erf((3/2+x)/eps)-erf((-3/2+x)/eps))/2;

% root node at level 0.
inode_now=1; % current node.
levelnodestart=zeros([ntotal_test(1)+1,1]); levelnodeend=levelnodestart;
levelnodestart(1)=1; % the starting node at level 0.
levelnodeend(1)=1; % the end node at level 0.
NODES(1,1) = 0; NODES(2,1) = 1; NODES(3,1) = 0; % root node
NODES(4,1) = 0; % no parent
NODES(5,1) = 0; NODES(6,1) = 0; % currently no children.

% generate the required NODES info for all levels.
for ilevel=1:ntotal_test(1)
    levelnodestart(ilevel+1)=inode_now+1; % the start node in this level.
    for iparent=levelnodestart(ilevel):levelnodeend(ilevel)
        % generate the two child nodes.
        inode_now=inode_now+1; ichild1=inode_now;
        inode_now=inode_now+1; ichild2=inode_now;
        
        NODES(5,iparent)=ichild1; NODES(6,iparent)=ichild2; %parent's children.
        
        NODES(1,ichild1)=ilevel; NODES(1,ichild2)=ilevel;
        NODES(2,ichild1)=ichild1-levelnodestart(ilevel+1)+1; 
        NODES(2,ichild2)=ichild2-levelnodestart(ilevel+1)+1;
        
        if NODES(3,iparent)==0
            NODES(3,ichild1)=1; NODES(3,ichild2)=2;
        elseif NODES(3,iparent)==1
            NODES(3,ichild1)=1; NODES(3,ichild2)=4;
        elseif NODES(3,iparent)==2
            NODES(3,ichild1)=3; NODES(3,ichild2)=2;
        else
            NODES(3,ichild1)=3; NODES(3,ichild2)=4;
        end
        
        NODES(4,ichild1)=iparent; NODES(4,ichild2)=iparent;
        NODES(5,ichild1)=0; NODES(6,ichild1)=0;
        NODES(5,ichild2)=0; NODES(6,ichild2)=0;
    end
    levelnodeend(ilevel+1)=inode_now;
end
%finished tree node generation.

% perform the upward pass recursively.
tic
compute_integral(1);
toc

COEFF(1,1,1)   % the computed integral.
